source("Library.R")
library(fda); library(mgcv);  library(pcaPP); library(chemometrics);
library(refund); library(lpSolve); library(gbm); library(glmnet); library(randomForest);
library(wavethresh); library(monomvn); library(nnls); library(ppls); library(mboost);
library(spikeSlabGAM); library(irlba); library(svcm); library(fda.usc)

library(fda)
library(fda.usc)
library(MASS)
library(refund)
library(wavethresh)
library(glmnet)
library(FRegSigCom)

data(Pork)
X=Pork$X
Y=Pork$Y
ntrain=80
ntest=nrow(X)-ntrain
xtrange=c(0,1) # the range of t in x(t).
t.x=seq(0,1,length.out=dim(X)[2])
t.size <- length(t.x)

nlevels <- 9
n.wv <- 2^nlevels
t.wv=seq(0,1,length.out=n.wv)
X.interpolate=matrix(0, nrow(X), n.wv)
for(i in 1:nrow(X))
{
  X.interpolate[i,]=approx(t.x, X[i,], t.wv)$y
}

diagmat <- diag(n.wv)
wave.mtx <- diagmat
for(i in 1:n.wv){
  tmp <- wd(diagmat[i,])
  tmp.cof <- accessC(tmp, level=0)
  for(j in 0:(nlevels-1))
    tmp.cof <- c(tmp.cof, accessD(tmp, level=j))
  wave.mtx[,i] <- tmp.cof
}

X.wv <- X.interpolate%*%t(wave.mtx)
for(nrep in 1:100)
{
  
  train.index=sample(1:dim(X)[1], ntrain)
  X.train <- X.test <- list()
  t.x.list=list()
  for(i in 1:1)
  {
    t.x.list[[i]]=t.x
    X.train[[i]]=X[train.index,]
    X.test[[i]]=X[-(train.index),]
  }
  Y.train <- Y[train.index]
  Y.test <- Y[-(train.index)]
  
  
  Xs.wv.train <- X.wv[train.index,]
  Xs.wv.test <- X.wv[-(train.index),]
  
  #######################################################
  
  # our method
  ptm <- proc.time()
  fit.cv6=cv.sof.spike(X.train, Y.train, t.x.list)
  Y.pred.save=pred.sof.spike(fit.cv6, X.test)
  pred.error.final=mean((Y.pred.save-Y.test)^2)
  our.pred.R2=1-sum((Y.pred.save-Y.test)^2)/(var(Y.test)*(ntest-1))
  Y.fit=pred.sof.spike(fit.cv6, X.train)
  our.R2=1-sum((Y.fit-Y.train)^2)/(var(Y.train)*(ntrain-1))
  time <-  (proc.time()-ptm)[3]
  
  print(c("pred.error.final=",pred.error.final))
  
  
  # refund: pfr.lf
  nbasis <- ntrain-1
  ptm <- proc.time()
  predCurve <- X[train.index,]
  pfr.lf <- pfr(Y.train~lf(predCurve, k=nbasis, bs="ps"))
  pred.pfr<- predict(pfr.lf, newdata=list(predCurve=X[-(train.index),]), type="response")
  pfr.pred.error <- mean((pred.pfr-Y.test)^2)
  pfr.pred.R2=1-sum((pred.pfr-Y.test)^2)/(var(Y.test)*(ntest-1))
  fit.pfr<- predict(pfr.lf, newdata=list(predCurve=X[(train.index),]), type="response")
  
  pfr.R2=1-sum((fit.pfr-Y.train)^2)/(var(Y.train)*(ntrain-1))
  print(c("pfr.pred.error=",pfr.pred.error))
  time <- c(time, (proc.time()-ptm)[3])
  
  
  ####################################################
  #fda.usc: fregre.pc
  ptm <- proc.time()
  X.fdata <- fdata(X, argvals=t.x)
  freg.pc <- fregre.pc.cv(X.fdata[train.index,], Y.train, lambda=TRUE, P=c(0,0,1))
  pred.freg.pc <- predict.fregre.fd(freg.pc$fregre.pc, X.fdata[-(train.index),])
  freg.pc.pred.error <- mean((pred.freg.pc-Y.test)^2)
  freg.pc.pred.R2 <- 1-sum((pred.freg.pc-Y.test)^2)/(var(Y.test)*(ntest-1))
  fit.freg.pc <- predict.fregre.fd(freg.pc$fregre.pc, X.fdata[(train.index),])
  
  freg.pc.R2 <- 1-sum((fit.freg.pc-Y.train)^2)/(var(Y.train)*(ntrain-1))
  time <- c(time, (proc.time()-ptm)[3])
  print(c("freg.pc.pred.error=",freg.pc.pred.error))
  
  
  #fregre.pls
  ptm <- proc.time()
  X.fdata <- fdata(X, argvals=t.x)
  freg.pls <- fregre.pls.cv(X.fdata[train.index,], Y.train, lambda=TRUE, P=c(0,0,1))
  pred.freg.pls <- predict.fregre.fd(freg.pls$fregre.pls, X.fdata[-(train.index),])
  freg.pls.pred.error <- mean((pred.freg.pls-Y.test)^2)
  freg.pls.pred.R2 <- 1-sum((pred.freg.pls-Y.test)^2)/(var(Y.test)*(ntest-1))
  fit.freg.pls <- predict.fregre.fd(freg.pls$fregre.pls, X.fdata[(train.index),])
  
  freg.pls.R2 <- 1-sum((fit.freg.pls-Y.train)^2)/(var(Y.train)*(ntrain-1))
  time <- c(time, (proc.time()-ptm)[3])
  print(c("freg.pls.pred.error=",freg.pls.pred.error))
  
  
  
  #############################################
  # wavelet + lasso or weighted lasso
  
  ptm <- proc.time()[3]
  wlasso.cv <- cv.glmnet(Xs.wv.train, Y.train, alpha=1, nfolds=10)
  wlasso.pred <- predict(wlasso.cv, Xs.wv.test, s="lambda.min")
  wlasso.pred.error <- mean((wlasso.pred-Y.test)^2)
  wlasso.pred.R2 <- 1-sum((wlasso.pred-Y.test)^2)/(var(Y.test)*(ntest-1))
  wlasso.fit <- predict(wlasso.cv, Xs.wv.train, s="lambda.min")
  wlasso.R2 <- 1-sum((wlasso.fit-Y.train)^2)/(var(Y.train)*(ntrain-1))
  time <- c(time, (proc.time()-ptm)[3])
  
  wt.v <- 1/as.numeric(apply(Xs.wv.train,2,var))
  wwlasso.cv <- cv.glmnet(Xs.wv.train, Y.train, alpha=1, nfolds=10, penalty.factor=wt.v)
  wwlasso.pred <-  predict(wwlasso.cv, Xs.wv.test, s="lambda.min")
  wwlasso.pred.error <- mean((wwlasso.pred-Y.test)^2)
  wwlasso.pred.R2 <- 1-sum((wwlasso.pred-Y.test)^2)/(var(Y.test)*(ntest-1))
  wwlasso.fit <- predict(wwlasso.cv, Xs.wv.train, s="lambda.min")
  wwlasso.R2 <- 1-sum((wwlasso.fit-Y.train)^2)/(var(Y.train)*(ntrain-1))
  
  time <- c(time, (proc.time()-ptm)[3])
  
  print("Lasso predict error")
  print(c(wlasso.pred.error, wwlasso.pred.error))
  
  #################################################
  #############################################
  # Ensemble methods
  ####################################################
  ptm <- proc.time()[3]
  use.methods <- c(
    "wrap.flm",  "wrap.pfr",  "wrap.flirti", "wrap.wnet", # FLMs
    "wrap.pclm", "wrap.fpcr", #PC-based FLMs
    "wrap.spikeslabgam", #PC-based FAM
    "wrap.sisr", "wrap.psr",  #Single-index models
    "wrap.fgam", "wrap.fregrenp", #non-linear/non-parametric
    "wrap.plsreg2", #PLS
    "wrap.lasso", "wrap.gbm", "wrap.rf") #lasso, boosted trees, rf
  
   
  set = list(spect=as.matrix(X),
             target=Y,
             name=paste("pork_ntrain=", ntrain,    sep=""))
  
  
  doOneRep <- function(spect, target, train, use.methods, J=10, PCA=FALSE){
    res <- superfunc(y.train=target[train], w.train=spect[train, , drop=FALSE],
                     w.valid=spect[-train, , drop=FALSE], methods=use.methods, J = J, PCA = PCA)
    
    res <- data.frame(t(c(colMeans((target[-train] - do.call(cbind, head(res, -2)))^2),
                          colMeans((target[-train] - res$Z.oos)^2))))
    
    return(res)
  }
  
  
  doOneSet <- function(name, spect, target, use.methods, reps=20, seed=NULL,
                       ntrain=floor(.7*length(target)), J=10, PCA=FALSE, cores=reps, useParallel=TRUE){
    stopifnot(length(target) == nrow(spect))
    
    if(useParallel) {
      stopifnot(require(parallel))
      options(mc.cores=cores)
      this.lapply <- function(X, FUN){
        mclapply(X=X, FUN=FUN, mc.preschedule = FALSE, mc.allow.recursive = FALSE)
      }
    } else this.lapply <- lapply
    
    if(is.null(seed)){
      seed <- runif(1, 0, 1e9)
    }
    
    n <- length(target)
    
    res <- this.lapply(1:reps, function(i){
      seed.i <- as.integer(runif(1, 0, 1e9))
      set.seed(seed.i)
      train <- train.index
      
      time.start = proc.time()
      
      tmp <- doOneRep(spect, target, train=train, use.methods, J=J, PCA=PCA)
      
      time.stop = proc.time()
      
      res <- cbind(tmp, name=name, rep=i, seed=seed.i, ntrain=ntrain,
                   benchmark.oos = mean((target[-train] - mean(target[train]))^2),
                   benchmark.is = mean((target[-train] - mean(target[-train]))^2),
                   time = (time.stop - time.start)[3])
      cat("\n", name,": replication ", i, " done.\n")
      return(res)
    })
    res <- do.call(rbind, res)
    
    return(res)
  }
  
  cores <- 1
  
  ret <- try(doOneSet(set$name, set$spect, set$target, use.methods, reps=1,
                      cores=cores,  ntrain=length(train.index), J=10, PCA=FALSE,
                      useParallel = FALSE))
  time <- c(time, (proc.time()-ptm)[3])
  
  #######################################################
  
  fileroot <- paste("pork_ntrain=", ntrain,    sep="")
  all <- c(pred.error.final,  freg.pls.pred.error, freg.pc.pred.error,   pfr.pred.error, wlasso.pred.error, wwlasso.pred.error, ret[1:6])
  write.table(t(all), file=paste(fileroot, ".error.txt",sep=""), append=TRUE,row.names=FALSE, col.names=FALSE)
}





