 
source("Library.R")
library(fda); library(mgcv);  library(pcaPP); library(chemometrics);
library(refund); library(lpSolve); library(gbm); library(glmnet); library(randomForest);
library(wavethresh); library(monomvn); library(nnls); library(ppls); library(mboost);
library(spikeSlabGAM); library(irlba); library(svcm); library(fda.usc)
#######################################################
# R code for simulation 1
############################################################
library(fda)
library(fda.usc)
library(MASS)
library(refund)
library(wavethresh)
library(glmnet)
library(FRegSigCom)



ncurves=1
ntot <- 600 # the total number of observations in both the training set and the test set
ntrain <- 100 # number of observations in training set
ntest <- ntot-ntrain # number of observations in test set

#######################################################
# the beta(t) in (5.1) used to generate the coefficient functions
############################################################
f.2=function(x)
{
  out=x
  y=x-floor(x)
  I=(y>=0)&(y<0.25)
  out[I]=4*y[I]
  I=(y>=0.25)&(y<0.75)
  out[I]=2-4*y[I]
  I=(y>=0.75)&(y<1)
  out[I]=4*y[I]-4
  return(out)
}



beta.1 <- function(t.x)
{
  y <- 0
  for(i in 2:65)
  {
    y <- y + (-1)^(floor(t.x*i) %% i)*(floor(t.x*i) %% i)
  }
  return(y)
}
beta.2 <- function(t.x)
{
  y <- 0
  for(j in 1:30)
  {
    i=j+50
    y=y+(f.2(i*(t.x)^0.5)+f.2(i*(t.x)^2)+f.2(i*(t.x)))/j^2
  }
  return(y)
}
beta.4 <- function(t.x)
{
  cos(10*pi*t.x)
}

gammaexpcov.1 <- function(n.point, t)
{
  tmp=sapply(1:length(t), function(i){abs(t-t[i])})
  Sigma=sapply(1:length(t), function(i){ exp(-3600*((t-t[i])^2))})
  Sigma <- (Sigma+t(Sigma))/2
  return(Sigma)
}

  


gammaexpcov.beta <- function(n.point, t)
{
  tmp=sapply(1:length(t), function(i){abs(t-t[i])})
  Sigma=sapply(1:length(t), function(i){ exp(-(400*abs(t-t[i]))^1.5)})
  Sigma <- (Sigma+t(Sigma))/2
  return(Sigma)
}


#######################################################
# 100 simulation iterations
############################################################
for(t.size in c(512))
{
  t.x=seq(0,1,length.out=t.size)
  Sigma.1  <- gammaexpcov.1(t.size, t.x)+1e-6*diag(t.size)
  R.1=chol(Sigma.1)
  
  Sigma.beta  <- gammaexpcov.beta(t.size, t.x)+1e-6*diag(t.size)
  R.beta=chol(Sigma.beta)
  nlevels <- floor(log2(t.size))
  n.wv <- 2^nlevels
  diagmat <- diag(n.wv)
  wave.mtx <- diagmat
  for(i in 1:n.wv){
    tmp <- wd(diagmat[i,])
    tmp.cof <- accessC(tmp, level=0)
    for(j in 0:(nlevels-1))
      tmp.cof <- c(tmp.cof, accessD(tmp, level=j))
    wave.mtx[,i] <- tmp.cof
  } # wv.coef.vec=W %*% x.vec
  
  for(nrep in 1:50)  {
    for(sigma.error in c(0.01,1))
    {
      for(ind.beta in 1:4)
      {
        for(x.type in c(1:2))
        {
          beta.val=list();
          beta.val[[1]] <- beta.1(t.x)
          beta.val[[2]] <- beta.2(t.x)
          t.x=seq(0,1,length.out=t.size)
          beta.val[[3]]=t(R.beta)%*%rnorm(t.size)
          beta.val[[4]] <- beta.4(t.x)
          # generate the predictor curves
          X <- matrix(0, ntot, t.size)
          if(x.type==1)
          {
            z=matrix(rnorm(t.size*ntot),ntot,t.size)
            X=t(crossprod(R.1, t(z)))
          }else
          {
            for(k in 1:30)
            {
              V.mtx <- matrix(rnorm(ntot*2,0,1/k), ncol=2)
              X <- X + V.mtx[,1] %*% t(sin(2*k*pi*t.x))+V.mtx[,2] %*% t(cos(2*k*pi*t.x)) 
            }
          }
          X.wv <- X%*%t(wave.mtx)
          
          # generate the regression function and response
          F=X%*%beta.val[[ind.beta]]/t.size
          beta=beta.val[[ind.beta]]/sqrt(mean(apply(F,2,var)))
          F=X%*%beta/t.size
          F=1+F # add the intercept to the matrix of regression function
          E <- sigma.error*rnorm(ntot) # noise matrix
          Y <- F + E # response matrix
          Y.train <-  Y[1:ntrain]
          F.test <-  F[-(1:ntrain)]
          X.train <- X.test <- list() # in our method, the observations of each predictor curve are need to be placed in a list
          t.x.list=list() # in our method, the observation time points of each predictor curve are need to be placed in a list
          for(i in 1:ncurves){
            t.x.list[[i]]=t.x
            X.train[[i]]=X[1:ntrain,]
            X.test[[i]]=X[-(1:ntrain),]
          }
          Xs.wv.train <- X.wv[1:ntrain,]
          Xs.wv.test <- X.wv[-(1:ntrain),]
          
          
          ###############################################################################
          ## analysis
          ##############################################################################
          #######################################################
          # our method
          #######################################################
          ptm <- proc.time()
          fit.cv=cv.sof.spike(X.train, Y.train, t.x.list)
          Y.pred=pred.sof.spike(fit.cv, X.test)
          pred.error.new=mean((Y.pred-F.test)^2)
          time <-  (proc.time()-ptm)[3]
          
          print(c("pred.error.new=",pred.error.new))
          
          
          #######################################################
          # pfr
          #######################################################
          nbasis <- 99
          ptm <- proc.time()
          predCurve <- X[1:ntrain,]
          pfr.lf <- pfr(Y.train~lf(predCurve, k=nbasis, bs="ps"))
          pred.pfr<- predict(pfr.lf, newdata=list(predCurve=X[-(1:ntrain),]), type="response")
          pfr.pred.error <- mean((pred.pfr-F.test)^2)
          print(c("pfr.pred.error=",pfr.pred.error))
          time <- c(time, (proc.time()-ptm)[3])
          
          ####################################################
          #fregre.pc
          ####################################################
          ptm <- proc.time()
          X.fdata <- fdata(X, argvals=t.x)
          freg.pc <- fregre.pc.cv(X.fdata[1:ntrain,], Y.train, lambda=TRUE, P=c(0,0,1))
          pred.freg.pc <- predict.fregre.fd(freg.pc$fregre.pc, X.fdata[-(1:ntrain),])
          freg.pc.pred.error <- mean((pred.freg.pc-F.test)^2)
          time <- c(time, (proc.time()-ptm)[3])
          print(c("freg.pc.pred.error=",freg.pc.pred.error))
          
          ####################################################
          #fregre.pls
          ####################################################
          ptm <- proc.time()
          X.fdata <- fdata(X, argvals=t.x)
          freg.pls <- fregre.pls.cv(X.fdata[1:ntrain,], Y.train, lambda=TRUE, P=c(0,0,1))
          pred.freg.pls <- predict.fregre.fd(freg.pls$fregre.pls, X.fdata[-(1:ntrain),])
          freg.pls.pred.error <- mean((pred.freg.pls-F.test)^2)
          time <- c(time, (proc.time()-ptm)[3])
          print(c("freg.pls.pred.error=",freg.pls.pred.error))
          
          #############################################
          # wavelet + lasso or weighted lasso
          ####################################################
          ptm <- proc.time()[3]
          wlasso.cv <- cv.glmnet(Xs.wv.train, Y.train, alpha=1, nfolds=10)
          wlasso.pred <- predict(wlasso.cv, Xs.wv.test, s="lambda.min")
          wlasso.pred.error <- mean((wlasso.pred-F.test)^2)
          time <- c(time, (proc.time()-ptm)[3])
          
          wt.v <- 1/as.numeric(apply(Xs.wv.train,2,var))
          wwlasso.cv <- cv.glmnet(Xs.wv.train, Y.train, alpha=1, nfolds=10, penalty.factor=wt.v)
          wwlasso.pred <-  predict(wwlasso.cv, Xs.wv.test, s="lambda.min")
          wwlasso.pred.error <- mean((wwlasso.pred-F.test)^2)
          time <- c(time, (proc.time()-ptm)[3])
          
          print("Lasso predict error")
          print(c(wlasso.pred.error, wwlasso.pred.error))
          
          #############################################
          # Ensemble methods
          ####################################################
          ptm <- proc.time()[3]
          use.methods <- c(
            "wrap.flm",  "wrap.pfr",  "wrap.flirti", "wrap.wnet", # FLMs
            "wrap.pclm", "wrap.fpcr", #PC-based FLMs
            "wrap.spikeslabgam", #PC-based FAM
            "wrap.sisr", "wrap.psr",  #Single-index models
            "wrap.fgam", "wrap.fregrenp", #non-linear/non-parametric
            "wrap.plsreg2", #PLS
            "wrap.lasso", "wrap.gbm", "wrap.rf") #lasso, boosted trees, rf
          
          
          Y[-(1:ntrain)] <- F[-(1:ntrain)]# use to calculate MSPE with respect to F
          set = list(spect=as.matrix(X),
                     target=Y,
                     name=paste("sim.1_sigma.error=", sigma.error, "_beta.type=", ind.beta, "_x.type=", x.type, ".mspe.txt", sep=""))
          
          
          doOneRep <- function(spect, target, train, use.methods, J=10, PCA=FALSE){
            res <- superfunc(y.train=target[train], w.train=spect[train, , drop=FALSE],
                             w.valid=spect[-train, , drop=FALSE], methods=use.methods, J = J, PCA = PCA)
            
            res <- data.frame(t(c(colMeans((target[-train] - do.call(cbind, head(res, -2)))^2),
                                  colMeans((target[-train] - res$Z.oos)^2))))
            
            return(res)
          }
          
          
          doOneSet <- function(name, spect, target, use.methods, reps=20, seed=NULL,
                               ntrain=floor(.7*length(target)), J=10, PCA=FALSE, cores=reps, useParallel=TRUE){
            stopifnot(length(target) == nrow(spect))
            
            if(useParallel) {
              stopifnot(require(parallel))
              options(mc.cores=cores)
              this.lapply <- function(X, FUN){
                mclapply(X=X, FUN=FUN, mc.preschedule = FALSE, mc.allow.recursive = FALSE)
              }
            } else this.lapply <- lapply
            
            if(is.null(seed)){
              seed <- runif(1, 0, 1e9)
            }
            
            n <- length(target)
            
            res <- this.lapply(1:reps, function(i){
              seed.i <- as.integer(runif(1, 0, 1e9))
              set.seed(seed.i)
              train <- 1:ntrain
              
              time.start = proc.time()
              
              tmp <- doOneRep(spect, target, train=train, use.methods, J=J, PCA=PCA)
              
              time.stop = proc.time()
              
              res <- cbind(tmp, name=name, rep=i, seed=seed.i, ntrain=ntrain,
                           benchmark.oos = mean((target[-train] - mean(target[train]))^2),
                           benchmark.is = mean((target[-train] - mean(target[-train]))^2),
                           time = (time.stop - time.start)[3])
              cat("\n", name,": replication ", i, " done.\n")
              return(res)
            })
            res <- do.call(rbind, res)
            
            return(res)
          }
          
          cores <- 1
          
          ret <- try(doOneSet(set$name, set$spect, set$target, use.methods, reps=1,
                              cores=cores,  ntrain=100, J=10, PCA=FALSE,
                              useParallel = FALSE))
          time <- c(time, (proc.time()-ptm)[3])
          #############################################
          # results output
          ####################################################
          fileroot <- paste("sim.1_sigma.error=", sigma.error, "_timepoints=", length(t.x), "_beta.type=", ind.beta,"_x.type=", x.type,  sep="")
          write.table(c(pred.error.new, pfr.pred.error, freg.pc.pred.error, freg.pls.pred.error, wlasso.pred.error, wwlasso.pred.error, ret[1:6]), 
                      file=paste(fileroot, ".error.txt",sep=""), append=TRUE,row.names=FALSE, col.names=FALSE)
        }
      }
    }
  }
}
