

#######################################################
# R code for simulation 2
############################################################
library(fda)
library(fda.usc)
library(MASS)
library(refund)
library(wavethresh)
library(glmnet)
library(FRegSigCom)

ncurves=4
ntot <- 600 # the total number of observations in both the training set and the test set
ntrain <- 100 # number of observations in training set
ntest <- ntot-ntrain # number of observations in test set

#######################################################
# the beta(t) in (5.1) used to generate the coefficient functions
############################################################
 
beta.21 <- function(t.x)
{
  exp(-40*(t.x-0.25)^2)+exp(-40*(t.x-0.75)^2)
}
beta.22 <- function(t.x)
{
  cos(4*pi*t.x)
}
beta.23 <- function(t.x)
{
  sin(5*pi*t.x)/(1+t.x^2)
}

beta.24 <- function(t.x)
{
  sqrt(t.x)
}

gammaexpcov.1 <- function(n.point, t)
{
  tmp=sapply(1:length(t), function(i){abs(t-t[i])})
  Sigma=sapply(1:length(t), function(i){ exp(-1600*((t-t[i])^2))})
  Sigma <- (Sigma+t(Sigma))/2
  return(Sigma)
}


#######################################################
# 100 simulation iterations
############################################################
for(t.size in c(1024))
{
  t.x=seq(0,1,length.out=t.size)
  Sigma.1  <- gammaexpcov.1(t.size, t.x)+1e-6*diag(t.size)
  R.1=chol(Sigma.1)
  
  nlevels <- floor(log2(t.size))
  n.wv <- 2^nlevels
  diagmat <- diag(n.wv)
  wave.mtx <- diagmat
  for(i in 1:n.wv){
    tmp <- wd(diagmat[i,])
    tmp.cof <- accessC(tmp, level=0)
    for(j in 0:(nlevels-1))
      tmp.cof <- c(tmp.cof, accessD(tmp, level=j))
    wave.mtx[,i] <- tmp.cof
  } # wv.coef.vec=W %*% x.vec
  
  for(nrep in 1:100)  {
    for(sigma.error in c(0.01, 0.1,1))
    {
      for(ind.beta in 1:2)
      {
        for(x.type in c(1:2))
        {
          
          beta.val=list();
          beta.val[[1]]=list()
          t.x=seq(0,1,length.out=t.size)
          for(k in 1:ncurves)
          {
            z=rnorm(t.size)
            tmp=matrix(0, length(t.x),length(t.x))
            phi=0.1^(5/t.size)
            tmp.1=sapply(1:length(t.x), function(k){phi^(k-1)})
            C=sapply(1:length(t.x),function(k){tmp[k:length(t.x),k]=tmp.1[1:(length(t.x)-k+1)]; return(tmp[,k])})
            beta.val[[1]][[k]]=crossprod(C, z)
          }
          beta.val[[2]]=list()
          beta.val[[2]][[1]] <- beta.21(t.x)
          beta.val[[2]][[2]] <- beta.22(t.x)
          beta.val[[2]][[3]] <- beta.23(t.x)
          beta.val[[2]][[4]] <- beta.24(t.x)
          
          
          
          
          # generate the predictor curves
          W=list()
          for(m in 1:(ncurves+2))
          {
            W[[m]] <- matrix(0, ntot, t.size)
            if(x.type==1)
            {
              z=matrix(rnorm(t.size*ntot),ntot,t.size)
              W[[m]]=t(crossprod(R.1, t(z)))
            }else
            {
              for(k in 1:15)
              {
                V.mtx <- matrix(rnorm(ntot*2,0,1/k), ncol=2)
                W[[m]] <- W[[m]]+V.mtx[,1] %*% t(sin(2*k*pi*t.x))+V.mtx[,2] %*% t(cos(2*k*pi*t.x)) 
              }
            }
            
          }
          X=list()
          X.wv=NULL
          for(k in 1:ncurves)
          {
            X[[k]] <-(W[[k]]+W[[k+1]]+W[[k+2]])/3
            X.wv <- cbind(X.wv, X[[k]]%*%t(wave.mtx))
          }
          
          # generate the regression function and response
          F=0
          for(k in 1:ncurves)
          {
            F=F+X[[k]]%*%beta.val[[ind.beta]][[k]]/t.size
          }
          
          
          F=F/sqrt(mean(apply(F,2,var)))
          F=1+F # add the intercept to the matrix of regression function
          E <- sigma.error*rnorm(ntot) # noise matrix
          Y <- F + E # response matrix
          
          Y.train <-  Y[1:ntrain]
          Y.test <-  Y[-(1:ntrain)]
          F.test <-  F[-(1:ntrain)]
          X.train <- X.test <- list() # in our method, the observations of each predictor curve are need to be placed in a list
          t.x.list=list() # in our method, the observation time points of each predictor curve are need to be placed in a list
          for(i in 1:ncurves){
            t.x.list[[i]]=t.x
            X.train[[i]]=X[[i]][1:ntrain,]
            X.test[[i]]=X[[i]][-(1:ntrain),]
          }
          Xs.wv.train <- X.wv[1:ntrain,]
          Xs.wv.test <- X.wv[-(1:ntrain),]
          
          #######################################################
          
          ptm <- proc.time()
          fit.cv=cv.sof.spike(X.train, Y.train, t.x.list)
          Y.pred=pred.sof.spike(fit.cv, X.test)
          pred.error.final=mean((Y.pred-F.test)^2)
          time <-  (proc.time()-ptm)[3]
          
          print(c("pred.error.final=",pred.error.final))
          
          
          
          
          
          #############################################
          # wavelet + lasso or weighted lasso
          
          ptm <- proc.time()[3]
          wlasso.cv <- cv.glmnet(Xs.wv.train, Y.train, alpha=1, nfolds=10)
          wlasso.pred <- predict(wlasso.cv, Xs.wv.test, s="lambda.min")
          wlasso.pred.error <- mean((wlasso.pred-F.test)^2)
          time <- c(time, (proc.time()-ptm)[3])
          
          wt.v <- 1/as.numeric(apply(Xs.wv.train,2,var))
          wwlasso.cv <- cv.glmnet(Xs.wv.train, Y.train, alpha=1, nfolds=10, penalty.factor=wt.v)
          wwlasso.pred <-  predict(wwlasso.cv, Xs.wv.test, s="lambda.min")
          wwlasso.pred.error <- mean((wwlasso.pred-F.test)^2)
          time <- c(time, (proc.time()-ptm)[3])
          
          print("Lasso predict error")
          print(c(wlasso.pred.error, wwlasso.pred.error))
          
          
          #######################################################
          
          fileroot <- paste("sim.2_sigma.error=", sigma.error, "_timepoints=", length(t.x), "_beta.type=", ind.beta,"_x.type=", x.type,  sep="")
          all <- c(pred.error.final,   wlasso.pred.error, wwlasso.pred.error, time)
          write.table(t(all), file=paste(fileroot, ".error.txt",sep=""), append=TRUE,row.names=FALSE, col.names=FALSE)
        }
      }
    }
  }
}




