

#######################################################
# R code for simulation 3
############################################################
library(fdapace)
library(wavethresh)
library(FRegSigCom)


ncurves=1
ntot <- 600 # the total number of observations in both the training set and the test set
ntrain <- 100 # number of observations in training set
ntest <- ntot-ntrain # number of observations in test set



f.2=function(x)
{
  out=x
  y=x-floor(x)
  I=(y>=0)&(y<0.25)
  out[I]=4*y[I]
  I=(y>=0.25)&(y<0.75)
  out[I]=2-4*y[I]
  I=(y>=0.75)&(y<1)
  out[I]=4*y[I]-4
  return(out)
}


beta.1 <- function(t.x, t.y)
{
  y <- 0
  for(i in 16:30)
  {
    y=y+f.2(i*t.x)%*%t(f.2(i*t.y))/(i-15)^2
  }
  return(y)
}


beta.2 <- function(t.x, t.y)
{
  y <- 0
  for(i in 21:30)
  {
    y <- y + (-1)^(floor(t.x*i) %% i)*(floor(t.x*i) %% i)%*%t((-1)^(floor(t.y*i) %% i)*(floor(t.y*i) %% i))/(i-20)
  }
  return(y)
}

beta.3 <- function(t.x, t.y)
{
  y <- exp(-100*(t.x-0.5)^2)%*%t(exp(-100*(t.y-0.5)^2))
  
  return(y)
}



gammaexpcov.1 <- function(n.point, t)
{
  tmp=sapply(1:length(t), function(i){abs(t-t[i])})
  Sigma=sapply(1:length(t), function(i){ exp(-3600*((t-t[i])^2))})
  Sigma <- (Sigma+t(Sigma))/2
  return(Sigma)
}


n.wv <- 9
diagmat <- diag(2^n.wv)
W.x <- diagmat
for(i in 1:2^n.wv){
  tmp <- wd(diagmat[i,])
  tmp.cof <- accessC(tmp, level=0)
  for(j in 0:(n.wv-1))
    tmp.cof <- c(tmp.cof, accessD(tmp, level=j))
  W.x[,i] <- tmp.cof
}

lengthX=2^n.wv
lengthY=500
t.x <- seq(0,1,length=lengthX)
t.y <- seq(0,1,length=lengthY)
Sigma.1  <- gammaexpcov.1(lengthX, t.x)+1e-6*diag(lengthX)
R.1=chol(Sigma.1)

print(c("start simulating data"))
beta.val=list();
beta.val[[1]] <- beta.1(t.x, t.y)
beta.val[[2]] <- beta.2(t.x, t.y)
beta.val[[3]] <- beta.3(t.x, t.y)
for(n.repeat in 1:100)
{
  for(sigma.error in c(0.01, 0.1, 1))
  {
    for(ind.beta in 1:3)
    {
      for(x.type in c(1:2))
      {
        # generate the predictor curves
        X <- matrix(0, ntot,lengthX)
        if(x.type==1)
        {
          z=matrix(rnorm(lengthX*ntot),ntot,lengthX)
          X=t(crossprod(R.1, t(z)))
        }else
        {
          for(k in 1:30)
          {
            V.mtx <- matrix(rnorm(ntot*2,0,1/k), ncol=2)
            X <- X + V.mtx[,1] %*% t(sin(2*k*pi*t.x))+V.mtx[,2] %*% t(cos(2*k*pi*t.x)) 
          }
        }
         
        
        F= tcrossprod(X, t(beta.val[[ind.beta]]))/lengthX
        F=F/sqrt(mean(apply(F,2,var)))
        
        E <- sigma.error*matrix(rnorm(ntot*lengthY), ntot, lengthY)
        
        Y <- F + E
        X.train.list=list()
        X.test.list=list()
        t.x.list=list()
        X.wv <- NULL
        for(i in 1:ncurves)
        {
          t.x.list[[i]]=t.x
          X.train.list[[i]]=X[1:ntrain,]
          X.test.list[[i]]=X[-(1:ntrain),]
          X.wv <- cbind(X.wv, X %*% t(W.x))
        }
        Y.train <- Y[1:ntrain, ]
        Y.test <- Y[-(1:ntrain), ]
        E.test <- E[-(1:ntrain),]
        F.test <-  F[-(1:ntrain),]
        X.wv.train <- X.wv[1:ntrain,]
        X.wv.test <- X.wv[-(1:ntrain),]
        
        
        #new method
        ptm <- proc.time()[3]
        fit.cv=cv.fof.spike(X.train.list, Y.train, t.x.list, t.y)
        Y.pred=pred.fof.spike(fit.cv, X.test.list)
        error.new=mean((Y.pred-F.test)^2)
        print(c("prediction error of new method=", error.new))
        times <- (proc.time()-ptm)[3]
        
        #smooth.Sigcom method 
        ptm <- proc.time()[3]
        fit.cv1=cv.sigcom(X.train.list, Y.train, t.x.list, t.y, s.n.basis=lengthX, t.n.basis=lengthY)
        Y.pred.1 <- pred.sigcom(fit.cv1, X.test.list)
        error.sigcom=mean((Y.pred.1-F.test)^2)
        print(c("Sigcom prediction error=", error.sigcom))
        times <- c(times, (proc.time()-ptm)[3])
        
        
        
        #FPCReg pace with AIC
        ptm <- proc.time()[3]
        timeX <- t(matrix(rep(t.x, ntot), length(t.x), ntot))
        timeY <- t(matrix(rep(t.y, ntot), length(t.y), ntot))
        dataFPCR <- list(X1=list(Ly=X, Lt=timeX), Y=list(Ly=Y, Lt=timeY))
        fpcr.out <- FPCReg(dataFPCR, method="AIC", isNewSub=c(rep(0,ntrain),rep(1,ntot-ntrain)))
        error.pace.aic <- mean((fpcr.out$predictY[-(1:ntrain),]-F.test)^2)
        print(c("predict error of PACE with AIC=", error.pace.aic))
        times <- c(times, (proc.time()-ptm)[3])
        
        #FPCReg pace with BIC
        ptm <- proc.time()[3]
        timeX <- t(matrix(rep(t.x, ntot), length(t.x), ntot))
        timeY <- t(matrix(rep(t.y, ntot), length(t.y), ntot))
        dataFPCR <- list(X1=list(Ly=X, Lt=timeX), Y=list(Ly=Y, Lt=timeY))
        fpcr.out <- FPCReg(dataFPCR, method="BIC", isNewSub=c(rep(0,ntrain),rep(1,ntot-ntrain)))
        error.pace.bic <- mean((fpcr.out$predictY[-(1:ntrain),]-F.test)^2)
        print(c("predict error of PACE with BIC=", error.pace.bic))
        times <- c(times, (proc.time()-ptm)[3])
        
        #FPCReg pace with FVE
        ptm <- proc.time()[3]
        timeX <- t(matrix(rep(t.x, ntot), length(t.x), ntot))
        timeY <- t(matrix(rep(t.y, ntot), length(t.y), ntot))
        dataFPCR <- list(X1=list(Ly=X, Lt=timeX), Y=list(Ly=Y, Lt=timeY))
        fpcr.out <- FPCReg(dataFPCR, method="FVE", isNewSub=c(rep(0,ntrain),rep(1,ntot-ntrain)))
        error.pace.fve <- mean((fpcr.out$predictY[-(1:ntrain),]-F.test)^2)
        print(c("predict error of PACE with FVE=", error.pace.fve))
        times <- c(times, (proc.time()-ptm)[3])
        
        
        #sparse.Sigcom method 
        ptm <- proc.time()[3]
        
        cv.fit <- cv.fof.wv(X.wv.train, Y.train, t.y)
        Y.pred=pred.fof.wv(cv.fit, X.wv.test)
        error.wave=mean((Y.pred-F.test)^2)
        print(c("predict error wv=", error.wave))
        times <- c(times, (proc.time()-ptm)[3]);
        
        
        
        test.error.all <- c( error.new,  error.sigcom, error.pace.aic, error.pace.bic, error.pace.fve, error.wave, times)
        file.root<- paste("sim.3_sigma.error=", sigma.error, "_timepoints=", length(t.x), "_beta.type=", ind.beta,"_x.type=", x.type,  sep="")
        write.table( t(test.error.all),file=paste(file.root, ".error.txt", sep=""), row.names=FALSE, col.names=FALSE, append=TRUE)
      }
    }
  }
}


